# coding: utf-8
import logging
import random
import string

from django.conf import settings
from django.contrib import messages
from django.contrib.auth import user_logged_out
from django.contrib.auth.models import User
from django.core.mail import send_mail, EmailMultiAlternatives
from django.db import transaction
from django.template import Context
from django.template.loader import get_template, render_to_string

from mailup.clients import MailUpClientSingleton
from mailup.exceptions import RecipientNotFoundException
from mailup.providers import MailUpComponentProvider

from jmb.core.middleware.thread_local import get_request

from . import exceptions
from .utils import sso_client
from acomea.contact.models import Contact, Group
from acomea.profilo.models import AcomeaProfilo, RichiestaAttivazioneCliente
from acomea.profilo.utils import create_acomeaprofilo_and_rac, get_or_create_contact


if settings.ACOMEA_SSO_SERVER:

    class SSOUserNotObfuscable(Exception):
        def __init__(self, message=None):
            message = message or u"Offuscamento non possibile, utente cliente di AcomeA o di Gimme5"
            super(SSOUserNotObfuscable, self).__init__(message)

    sso = sso_client.SSOClient()

    ###################################################
    # Functions redefined to synch sso
    ###################################################
    @transaction.commit_manually
    def set_password(user, raw_password):
        try:
            original_set_password(user, raw_password)
            original_save(user)
            sso.admin_set_password(user.username, raw_password)
            transaction.commit()
        except Exception:
            transaction.rollback()
            raise

    @transaction.commit_manually
    def save(user, *args, **kwargs):
        try:
            create = not bool(user.id)
            sso_instance = None

            if not create:
                sso_instance = sso.detail(user.username)
                if sso_instance.get('data').get('ar') == 'C' or sso_instance.get('data').get('g5') == 'C':
                    sso_data = sso_instance.get('data')
                    if user.first_name != sso_data.get('cn') or user.last_name != sso_data.get('sn'):
                        log = logging.getLogger("acomea.utenti")
                        msg = u"Utente già cliente, Nome e cognome ripristinati con SSO; pk: {}, username: {}".format(
                                user.pk, user.username
                            )
                        log.warning(msg)
                        set_admin_warning(msg)
                        user.first_name = sso_data.get('cn')
                        user.last_name = sso_data.get('sn')
                        user.is_active = sso_data.get('disabled') == '0'

            # Sul db salvo ugualmente nel caso siano stati variati attributi non presenti in sso
            original_save(user, *args, **kwargs)

            if not create:
                # sso_instance = sso.detail(user.username)
                new_payload = dict()
                if user.email != sso_instance.get('data').get('mail'):
                    new_payload['email'] = user.email
                if user.first_name != sso_instance.get('data').get('cn'):
                    new_payload['first_name'] = user.first_name
                if user.last_name != sso_instance.get('data').get('sn'):
                    new_payload['last_name'] = user.last_name

                disabled = int(not user.is_active)
                if disabled != int(sso_instance.get('data').get('disabled')):
                    new_payload['disabled'] = disabled

                if new_payload:
                    sso.update(
                        username=user.username,
                        **new_payload
                    )
            transaction.commit()
        except Exception:
            transaction.rollback()
            raise

    @transaction.commit_manually
    def create(**kwargs):
        try:
            # Sul db salvo ugualmente nel caso siano stati variati attributi non presenti in sso
            user = original_create(**kwargs)

            payload = {
                "username": user.username,
                "acc_trattamento_dati_int": None,
                "acc_profilazione_int": None,
                "password": kwargs.get('password'),
                "ar": None,
                "codice_fiscale": None,
                "first_name": user.first_name,
                'g5': None,
                "email": user.email,
                "last_name": user.last_name,
                "disabled": int(not user.is_active)
            }
            sso.register(**payload)
            transaction.commit()
            return user
        except Exception:
            transaction.rollback()
            raise

    # monkey patch for replace original delete with obfuscate (#REPLACE_DELETE_OBFUSCATE)
    def obfuscate(user, *args, **kwargs):

        email = user.email
        # uso u'..' per controllo su ldap
        user.first_name = u'deleted'
        user.last_name = u'deleted'
        # username non cambia per compatibilità con Gimme5
        # self.username = u'deleted_{}'.format(self.pk)
        user.email = u'deleted_{}@deleted.com'.format(user.pk)
        user.is_active = False
        try:
            sso_instance = sso.detail(user.username)
            if sso_instance.get('data').get('ar') == 'C' or sso_instance.get('data').get('g5') == 'C':
                raise SSOUserNotObfuscable()
            user.save()
            sso.disable_user(user.username)
        except SSOUserNotObfuscable:
            raise
        except Exception:
            original_save(user, *args, **kwargs)
            send_email_user_not_exist_on_sso(user)

        if hasattr(user, 'contact'):
            user.contact.obfuscate()

        if hasattr(user, 'profilo'):
            user.profilo.obfuscate(notify_user_not_exist_on_sso=False)

            if hasattr(user.profilo, 'rac'):
                user.profilo.rac.obfuscate(notify_user_not_exist_on_sso=False)

        mailup_client = MailUpClientSingleton(
            client_id=settings.ACOMEA_MAILUP_CLIENT_ID,
            client_secret=settings.ACOMEA_MAILUP_SECRET,
            username=settings.ACOMEA_MAILUP_USERNAME,
            password=settings.ACOMEA_MAILUP_PASSWORD,
        )

        provider = MailUpComponentProvider(client=mailup_client)
        try:
            recipient = provider.get_recipient(settings.ACOMEA_MAILUP_LIST_ID, email=email)
            recipient.set_fields({
                'Nome': user.first_name,
                'Cognome': user.last_name,
                'ID_USER': user.username
            })
            recipient.email = user.email
            recipient.save()

            recipient.unsubscribe_to_list(settings.ACOMEA_MAILUP_LIST_ID)
        except RecipientNotFoundException:
            log = logging.getLogger("acomea.utenti")
            log.warning("OBFUSCATE The recipient {email} not found in MAIL_UP".format(email=email))

    def user_logout(sender, user, request, **kwargs):
        if not user:
            # in caso di refresh della pagina con url di logout (logout doppio), lo user è None
            return None
        try:
            sso.logout(
                username=user.username,
                token=request.COOKIES.get('token')
            )
        except Exception:
            return None

    @transaction.commit_manually
    def delete(user, *args, **kwargs):
        try:
            original_delete(user, *args, **kwargs)
            transaction.commit()
        except Exception:
            transaction.rollback()
            raise

    @transaction.commit_manually
    def profilo_save(profilo, *args, **kwargs):
        try:
            # Sul db salvo ugualmente nel caso siano stati variati attributi non presenti in sso
            profilo_original_save(profilo, *args, **kwargs)

            sso_instance = sso.detail(profilo.user.username)
            ar = 'C' if profilo.is_cliente else 'L'
            new_payload = dict()
            if ar != sso_instance.get('data').get('ar'):
                new_payload['ar'] = ar
            # acc_trattamento_dati_int = int(profilo.auth_trattamento_dati)
            # if str(acc_trattamento_dati_int) != sso_instance.get('data').get('accTrattamentoDatiInt'):
            #     new_payload['acc_trattamento_dati_int'] = acc_trattamento_dati_int
            if new_payload:
                sso.update(
                    username=profilo.user.username,
                    **new_payload
                )
            transaction.commit()
        except Exception:
            transaction.rollback()
            raise

    # monkey patch for replace original delete with obfuscate (#REPLACE_DELETE_OBFUSCATE)
    def profilo_obfuscate(profilo, notify_user_not_exist_on_sso=True, *args, **kwargs):
        try:
            sso_instance = sso.detail(username=profilo.user.username)
            if sso_instance.get('data').get('ar') == 'C' or sso_instance.get('data').get('g5') == 'C':
                raise SSOUserNotObfuscable()
            profilo_original_obfuscate(profilo)
        except SSOUserNotObfuscable:
            raise
        except Exception:
            profilo = profilo_original_obfuscate(profilo, to_save=False)
            profilo_original_save(profilo, *args, **kwargs)
            if notify_user_not_exist_on_sso:
                send_email_user_not_exist_on_sso(profilo)

    @transaction.commit_manually
    def profilo_create(**kwargs):
        try:
            profilo = profilo_original_create(**kwargs)

            sso_instance = sso.detail(profilo.user.username)
            ar = 'C' if profilo.is_cliente else 'L'
            new_payload = dict()
            if ar != sso_instance.get('data').get('ar'):
                new_payload['ar'] = ar
            # acc_trattamento_dati_int = int(profilo.auth_trattamento_dati)
            # if str(acc_trattamento_dati_int) != sso_instance.get('data').get('accTrattamentoDatiInt'):
            #     new_payload['acc_trattamento_dati_int'] = acc_trattamento_dati_int
            if new_payload:
                sso.update(
                    username=profilo.user.username,
                    **new_payload
                )
            transaction.commit()
            return profilo
        except Exception:
            transaction.rollback()
            raise

    @transaction.commit_manually
    def rac_save(rac, *args, **kwargs):
        try:
            sso_instance = sso.detail(rac.profilo.user.username)
            if sso_instance.get('data').get('ar') == 'C' or sso_instance.get('data').get('g5') == 'C':
                if rac.codice_fiscale != sso_instance.get('data').get('cf'):
                    log = logging.getLogger("acomea.utenti")
                    msg = u"Utente già cliente, codice fiscale ripristinato con SSO; pk: {}, username: {}".format(
                            rac.profilo.user.pk, rac.profilo.user.username
                        )
                    log.warning(msg)
                    set_admin_warning(msg)
                    rac.codice_fiscale = sso_instance.get('data').get('cf')

            rac_original_save(rac, *args, **kwargs)

            new_payload = dict()
            if rac.codice_fiscale != sso_instance.get('data').get('cf'):
                if rac.codice_fiscale and rac.codice_fiscale.lower() == 'undefined':
                    rac.codice_fiscale = rac.codice_fiscale.lower()
                new_payload['codice_fiscale'] = rac.codice_fiscale
                sso.update(
                    username=rac.profilo.user.username,
                    **new_payload
                )

            transaction.commit()
        except Exception:
            transaction.rollback()
            raise

    def rac_obfuscate(rac, notify_user_not_exist_on_sso=True, *args, **kwargs):
        try:
            sso_instance = sso.detail(username=rac.profilo.user.username)
            if sso_instance.get('data').get('ar') == 'C' or sso_instance.get('data').get('g5') == 'C':
                raise SSOUserNotObfuscable()
            rac_original_obfuscate(rac)
        except SSOUserNotObfuscable:
            raise
        except Exception:
            rac = rac_original_obfuscate(rac, to_save=False)
            rac_original_save(rac, *args, **kwargs)
            if notify_user_not_exist_on_sso:
                send_email_user_not_exist_on_sso(rac)

    @transaction.commit_manually
    def rac_create(**kwargs):
        try:
            new_payload = dict()
            rac = rac_original_create(**kwargs)

            sso_instance = sso.detail(rac.profilo.user.username)
            if rac.codice_fiscale != sso_instance.get('data').get('cf'):
                new_payload['codice_fiscale'] = rac.codice_fiscale
                sso.update(
                    username=rac.profilo.user.username,
                    **new_payload
                )
            transaction.commit()
            return rac
        except Exception:
            transaction.rollback()
            raise

    @transaction.commit_manually
    def contact_save(contact, *args, **kwargs):
        try:
            # Sul db salvo ugualmente nel caso siano stati variati attributi non presenti in sso
            contact_original_save(contact, *args, **kwargs)
            if hasattr(contact, 'user') and contact.user:
                try:
                    sso.exist(username=contact.user.username)
                except Exception:
                    raise exceptions.UserNotFoundOnSSOException(u'username {} not found on sso')

                sso_instance = sso.detail(contact.user.username)
                new_payload = dict()
                acc_trattamento_dati_int = int(contact.auth_trattamento_dati)
                if str(acc_trattamento_dati_int) != sso_instance.get('data').get('accTrattamentoDatiInt'):
                    new_payload['acc_trattamento_dati_int'] = acc_trattamento_dati_int
                acc_profilazione_int = int(contact.auth_profilazione)
                if str(acc_profilazione_int) != sso_instance.get('data').get('accProfilazioneDatiInt'):
                    new_payload['acc_profilazione_int'] = acc_profilazione_int
                if new_payload:
                    sso.update(
                        username=contact.user.username,
                        **new_payload
                    )
            transaction.commit()

        except exceptions.UserNotFoundOnSSOException:
            # committo modifica su db anche se utente non presente su sso
            transaction.commit()
            raise

        except Exception:
            transaction.rollback()
            raise

    def contact_obfuscate(contact, notify_user_not_exist_on_sso=True, *args, **kwargs):
        try:
            sso_instance = sso.detail(username=contact.user.username)
            if sso_instance.get('data').get('ar') == 'C' or sso_instance.get('data').get('g5') == 'C':
                raise SSOUserNotObfuscable()
            contact_original_obfuscate(contact)
        except SSOUserNotObfuscable:
            raise
        except Exception:
            contact = contact_original_obfuscate(contact, to_save=False)
            contact_original_save(contact, *args, **kwargs)
            if notify_user_not_exist_on_sso:
                send_email_user_not_exist_on_sso(contact)

    # monkey patch for replace original delete with obfuscate (#REPLACE_DELETE_OBFUSCATE)

    # USER
    original_set_password = User.set_password
    User.set_password = set_password
    original_save = User.save
    User.save = save
    original_create = User.objects.create
    User.objects.create = create
    original_obfuscate = User.obfuscate
    User.obfuscate = obfuscate
    original_delete = User.delete
    User.delete = obfuscate
    # USER SIGNAL
    user_logged_out.connect(user_logout)

    # PROFILO
    profilo_original_save = AcomeaProfilo.save
    AcomeaProfilo.save = profilo_save
    profilo_original_obfuscate = AcomeaProfilo.obfuscate
    AcomeaProfilo.obfuscate = profilo_obfuscate
    profilo_original_create = AcomeaProfilo.objects.create
    AcomeaProfilo.objects.create = profilo_create

    # RAC
    rac_original_save = RichiestaAttivazioneCliente.save
    RichiestaAttivazioneCliente.save = rac_save
    rac_original_obfuscate = RichiestaAttivazioneCliente.obfuscate
    RichiestaAttivazioneCliente.obfuscate = rac_obfuscate
    rac_original_create = RichiestaAttivazioneCliente.objects.create
    RichiestaAttivazioneCliente.objects.create = rac_create

    # CONTACT
    contact_original_save = Contact.save
    Contact.save = contact_save
    contact_original_obfuscate = Contact.obfuscate
    Contact.obfuscate = contact_obfuscate

    ###################################################
    # Functions called from auth flow
    ###################################################

    def generate_unique_username(base_string, count=0):
        username = '{}'.format(base_string)
        if count:
            username = '{}{}'.format(base_string, count)
        try:
            sso.exist(username=username)
            return generate_unique_username(base_string, count + 1)
        except Exception:
            return username

    def generate_random_string(len_string=8):
        allowed_chars = ''.join((string.digits, string.ascii_letters))
        return ''.join([random.choice(allowed_chars) for i in range(len_string)])

    def create_dj_user(data, password):
        dj_user = User(
            username=data.get('uid'),
            email=data.get('mail'),
            password=password,
            first_name=data.get('cn'),
            last_name=data.get('sn'),
            is_superuser=False,
        )
        # dj_user.set_password(password)
        # original_set_password(dj_user, password)
        # dj_user.save()
        original_save(dj_user)
        return dj_user

    def update_dj_user(dj_user, data, password):
        dj_user.email = data.get('mail')
        dj_user.first_name = data.get('cn')
        dj_user.last_name = data.get('sn')
        # dj_user.set_password(password)
        if password:
            original_set_password(dj_user, password)
        # dj_user.save()
        original_save(dj_user)
        return dj_user

    def update_profilo(dj_user, data):
        if hasattr(dj_user, 'profilo') and hasattr(dj_user.profilo, 'rac'):
            if(
                not dj_user.profilo.is_cliente or not dj_user.profilo.is_promotore or
                not dj_user.profilo.is_utente_rete or not dj_user.profilo.is_acomea
            ):
                # dj_user.profilo.rac.codice_fiscale = data.get('cf')
                rac = dj_user.profilo.rac
                rac.codice_fiscale = data.get('cf')
                rac_original_save(rac)

    def refresh_or_create_dj_objects(sso_login_data, password=None):
        # se password==None, arrivo da middleware SingleSignOn:
        #   se ho l'utente in dj non cambio la pwd
        #   se non ho l'utente in dj creo una pwd random
        sso_username = sso_login_data.get('uid')
        data = sso.detail(sso_username).get('data')
        try:
            dj_user = User.objects.get(username=sso_username)
            update_dj_user(dj_user, data, password)
            update_profilo(dj_user, data)
            get_or_create_contact(dj_user)  # last modifier?
            # servirà nel logout, sso logout vuole token
            setattr(dj_user, 'sso', sso_login_data)
            return dj_user
        except User.DoesNotExist:
            try:
                dj_user = User.objects.get(email=data.get('mail'))
                send_email_new_user_same_email_exist_from_sso(data, dj_user)
                raise exceptions.DuplicateEmailIntegrityException('Email già presente in Acomea.')
            except User.DoesNotExist:
                password = settings.ACOMEA_SSO_PWD_DEFAULT
                dj_user = create_dj_user(data, password)
                create_acomeaprofilo_and_rac(dj_user)
                get_or_create_contact(dj_user)  # last modifier?
                group = Group.objects.get(id=settings.ACOMEA_GROUP_FROM_G5)
                if group:
                    dj_user.contact.groups.add(group)
                # servirà nel logout, sso logout vuole token
                setattr(dj_user, 'sso', sso_login_data)
                return dj_user

    def send_email_user_not_exist_on_sso(instance):
        pk = instance.pk
        instance_name = instance.__class__.__name__
        try:
            # User
            username = instance.username
            email = instance.email
        except Exception:
            # Profilo/Contact
            try:
                username = instance.user.username
                email = instance.user.email
            except Exception:
                # Rac
                try:
                    username = instance.profilo.user.username
                    email = instance.profilo.user.email
                except Exception:
                    username = ''
                    email = ''

        email_context = Context({
            'obj': instance_name,
            'pk': pk,
            'username': username,
            'email': email
        })

        text_template = 'sso/email/user_not_exist_on_sso.{}'.format('txt')
        text_message = render_to_string(text_template, email_context)

        html_template = 'sso/email/user_not_exist_on_sso.{}'.format('html')
        html_message = render_to_string(html_template, email_context)

        msg = EmailMultiAlternatives(
            subject='Obfuscate error on SSO (user {})'.format(username),
            body=text_message,
            from_email=settings.ACOMEA_SSO_NO_REPLY_EMAIL,
            to=settings.ACOMEA_SSO_OBFUSCATE_ERROR_EMAIL_TO
        )
        msg.attach_alternative(html_message, "text/html")
        msg.send()

    def send_email_new_user_same_email_exist_from_sso(sso_data, dj_user):

        email_context = Context({
            'sso': {'username': sso_data.get('uid'), 'email': sso_data.get('mail')},
            'dj': dj_user,
        })

        text_template = 'sso/email/new_user_same_email_exist_from_sso.{}'.format('txt')
        text_message = render_to_string(text_template, email_context)

        html_template = 'sso/email/new_user_same_email_exist_from_sso.{}'.format('html')
        html_message = render_to_string(html_template, email_context)

        msg = EmailMultiAlternatives(
            subject='Errore accesso utente sso (user {})'.format(sso_data.get('uid')),
            body=text_message,
            from_email=settings.ACOMEA_SSO_NO_REPLY_EMAIL,
            to=settings.ACOMEA_SSO_OBFUSCATE_ERROR_EMAIL_TO
        )
        msg.attach_alternative(html_message, "text/html")
        msg.send()

    def set_admin_warning(msg):
        request = get_request()
        if request:
            messages.add_message(request, messages.WARNING, msg)
