# -*- coding: utf-8 -*-
import json
import logging

import requests

from six import string_types

from django.conf import settings
from django.utils.text import ugettext_lazy as _


class SSOClient:

    logger = logging.getLogger('acomea')

    def __init__(self):
        self.sso_auth = settings.ACOMEA_SSO_OAUTH
        self.headers = {
            'content-type': "application/json",
            'x-api-key': self.sso_auth['SECRET_KEY']
        }

    def call_sso_endpoint(self, action, payload_as_dict, get_response=False):
        url = self.get_sso_endpoint(action)
        payload_as_dict["srcSys"] = self.sso_auth['SRC_SYS']
        payload = json.dumps(self.clean_payload(payload_as_dict))
        # self.logger.info('{}\t{}'.format(action, payload))
        response = requests.request("POST", url, data=payload, headers=self.headers)
        if get_response:
            return response
        if response.status_code == 200:
            return response.json()
        # print (url, payload)
        # print (response.status_code, response.json())

        raise Exception("{}: {}".format(response.json().get('status'), response.json().get('message')))

    def clean_payload(self, payload):
        new_payload = dict()
        for key, value in payload.items():
            if value is None:
                continue
            if isinstance(value, string_types) and value.strip() == '':
                continue
            new_payload[key] = u'{}'.format(value)
        return new_payload

    def get_sso_endpoint(self, action):
        return "{}{}".format(self.sso_auth['BASE_ENDPOINT'], action)

    def register(
        self, username, password, acc_trattamento_dati_int, acc_profilazione_int,
        ar, codice_fiscale, first_name, g5, email, last_name, disabled
    ):
        payload = {
            'username': username,
            'password': password,
            'accTrattamentoDatiInt': acc_trattamento_dati_int,
            'accProfilazioneDatiInt': acc_profilazione_int,
            'ar': ar,
            'cf': codice_fiscale,
            'cn': first_name,
            'g5': g5,
            'mail': email,
            'sn': last_name,
            'disabled': disabled
        }
        return self.call_sso_endpoint('register', payload)

    def register_social(
        self, username, password, acc_trattamento_dati_int, acc_profilazione_int, ar, codice_fiscale,
        first_name, g5, email, last_name, social_unique_id, social_type, disabled
    ):
        payload = {
            'username': username,
            'password': password,
            'accTrattamentoDatiInt': acc_trattamento_dati_int,
            'accProfilazioneDatiInt': acc_profilazione_int,
            'ar': ar,
            'cf': codice_fiscale,
            'cn': first_name,
            'g5': g5,
            'mail': email,
            'sn': last_name,
            'socialUniqueID': social_unique_id,
            'socialType': social_type,
            'disabled': disabled
        }
        return self.call_sso_endpoint('register', payload)

    def insert(self, payload_as_dict):
        # usata da signals pre save: 'import' accetta la pwd dell'istanza già criptata;
        # 'register' vuole la pwd in chiaro
        return self.call_sso_endpoint('import', payload_as_dict)

    def update(
        self, username, acc_trattamento_dati_int=None, acc_profilazione_int=None,
        ar=None, codice_fiscale=None, first_name=None, g5=None, email=None, last_name=None, disabled=None
    ):
        payload = {
            'username': username,
            'accTrattamentoDatiInt': acc_trattamento_dati_int,
            'accProfilazioneDatiInt': acc_profilazione_int,
            'ar': ar,
            'cf': codice_fiscale,
            'cn': first_name,
            'g5': g5,
            'mail': email,
            'sn': last_name,
            # probabilmente non è mai stato inserito perchè se l'utente dovesse essere disabilitato su SSO
            # e non è ancora girato lo script di allineamento, erroneamente si potrebbe riabilitare un utente
            # 'disabled': disabled
        }
        return self.call_sso_endpoint('update', payload)

    def check(self, username, token):
        payload = {
            'username': username,
            'token': token
        }
        resp = self.call_sso_endpoint('check', payload)
        if settings.ACOMEA_SSO_AUTH_VERIFY_DISABLE:
            detail = self.detail(username=username)
            if detail.get('data', {}).get('disabled') == '1':
                raise Exception("{}: {}".format(401, _('Token valid but user disabled')))
        return resp

    def login(self, username, password):
        payload = {
            'username': username,
            'password': password
        }
        resp = self.call_sso_endpoint('login', payload)
        if settings.ACOMEA_SSO_AUTH_VERIFY_DISABLE:
            sso_username = resp.get('uid')
            detail = self.detail(username=sso_username)
            if detail.get('data', {}).get('disabled') == '1':
                raise Exception("{}: {}".format(401, _('Token valid but user disabled')))
        return resp

    def login_social(self, social_unique_id, social_type):
        payload = {
            'socialUniqueID': social_unique_id,
            'socialType': social_type
        }
        return self.call_sso_endpoint('login', payload)

    def logout(self, username, token):
        payload = {
            'username': username,
            'token': token
        }
        return self.call_sso_endpoint('logout', payload)

    def change_password(self, username, password, new_password):
        payload = {
            'username': username,
            'password': password,
            'new_password': new_password
        }
        return self.call_sso_endpoint('changepassword', payload)

    def forget_password(self, username):
        payload = {
            'username': username,
            'otp': 'request'
        }
        return self.call_sso_endpoint('forgetpassword', payload)

    def set_password(self, username, otp_pass, password):
        payload = {
            "username": username,
            "otp": "check",
            "otp_pass": otp_pass,
            "new_pass": password,
        }
        return self.call_sso_endpoint('forgetpassword', payload)

    def admin_set_password(self, username, password):
        data = self.forget_password(username)
        return self.set_password(username, data.get('otp_code'), password)

    def bulk_import(
        self, username, hash_password, acc_trattamento_dati_int,
        ar, codice_fiscale, first_name, g5, email, last_name, disabled,
    ):
        payload = {
            'username': username,
            'password': hash_password,
            'accTrattamentoDatiInt': acc_trattamento_dati_int,
            'ar': ar,
            'cf': codice_fiscale,
            'cn': first_name,
            'g5': g5,
            'mail': email,
            'sn': last_name,
            'disabled': disabled,
        }
        return self.call_sso_endpoint('import', payload)

    def detail(self, username=None, email=None, codice_fiscale=None, get_response=False):
        payload = {
            'username': username,
            'mail': email,
            'cf': codice_fiscale,
        }
        return self.call_sso_endpoint('list', payload, get_response=get_response)

    def exist(self, username=None, codice_fiscale=None, email=None, get_response=False):
        payload = {
            'username': username,
            'cf': codice_fiscale,
            'mail': email
        }
        return self.call_sso_endpoint('exist', payload, get_response=get_response)

    def manage_social(self, username, social_unique_id, social_type, action):
        payload = {
            'username': username,
            'socialUniqueID': social_unique_id,
            'socialType': social_type,  # facebook | google
            'action': action,  # associate | disassociate | forceassociate
        }
        return self.call_sso_endpoint('managesocial', payload)

    def social_token(self, username):
        payload = {
            'username': username,
        }
        resp = self.call_sso_endpoint('socialtoken', payload)
        if settings.ACOMEA_SSO_AUTH_VERIFY_DISABLE:
            sso_username = resp.get('uid')
            detail = self.detail(username=sso_username)
            if detail.get('data', {}).get('disabled') == '1':
                raise Exception("{}: {}".format(401, _('Token valid but user disabled')))
        return resp

    def enable_user(self, username):
        payload = {
            'username': username,
            'disabled': '0'
        }
        return self.call_sso_endpoint('update', payload)

    def disable_user(self, username):
        payload = {
            'username': username,
            'disabled': '1'
        }
        return self.call_sso_endpoint('update', payload)

